/*------------------------------------------------------------------------------
 *
 *      PicAKE SM6LKM's "Pic AccuKeyer Emulator"
 *
 *      Project:
 *        A keyer that emulates WB4VVF's legendary "Accu-Keyer" which is
 *        a two-paddle "Mode B" iambic keyer with automatic character space.
 *
 *      Processor:
 *        PIC12F675
 *
 *      Clock:
 *        4 MHz INTOSC
 *
 *      Compiler:
 *        BKD CC5X v3.1j Extended
 *
 *      Dependencies:
 *        12f675.h (supplied with compiler)
 *        int16cxx.h (supplied with compiler)
 *
 *      Notes:
 *        This is *not* a low power version, the CPU runs continously and
 *        the current gulping brown-out detector is enabled. (A QRP version
 *        would use CPU sleep and switched power to the speed potentiometer
 *        etc; this is left as an excercise for the reader ;-)
 *
 *      - Revision history -
 *      ========================================================================
 *      Version  Date    Comment / Programmer
 *      ------------------------------------------------------------------------
 *      1.00     060626  Created, based on experimental qrpk675.c / JHB SM6LKM
 *      ========================================================================
 */

#include <12f675.h>
#include <int16cxx.h>


/*------------------------------------------------------------------------------
 *
 *      Insert new oscillator calibration value (for development only)
 *
 *      Use #defined constant below or use command line option -D.
 *
 *      Example:
 *        CC5X <filename.c> -DCALVAL=0x??
 */

//#define CALVAL 0xA0                     // *#* hardcoded if #defined!!!

#ifdef CALVAL                           // calibration data on command line?
#define CALVAL_ADDRESS  0x3FF           // where to put RETLW CALVAL
#define RETLW_INSTR     0x3400
#pragma cdata[CALVAL_ADDRESS]= RETLW_INSTR + CALVAL
#endif


/*------------------------------------------------------------------------------
 *
 *      Set configuration word
 *
 *      Bits (PIC12F630/675):
 *        Gg  = BG1:BG0, Bandgap calibration bits
 *                00 = Lowest bandgap voltage
 *                ...
 *                11 = Highest bandgap voltage
 *        -   = Unimplemented, read as 0
 *        D   = /CPD, Data EEPROM protection, 0=protected, 1=unprotected
 *        C   = /CP, Code memory protection, 0=protected, 1=unprotected
 *        B   = BODEN, Brown-out detector, 0=disabled, 1=enabled
 *        M   = MCLRE, GP3 function select, 0=GP3 is input, 1=GP3 pin is /MCLR
 *        P   = /PWRTE, Power-up timer, 0=enabled, 1=disabled
 *        W   = WDTE, Watchdog timer, 0=disabled, 1=enabled
 *        XYZ = FOSC2:FOSC0, Oscillator mode
 *               111 = RC, GP4 is CLKOUT, GP5 connected to external RC
 *               110 = RC, GP4 is I/O, GP5 connected to external RC
 *               101 = INTOSC, GP4 is CLKOUT, GP5 is I/O
 *               100 = INTOSC, GP4 is I/O, GP5 is I/O
 *               011 = EC, GP4 is I/O, GP5 is CLKIN
 *               010 = HS, crystal or resonator connected to GP4 and GP5
 *               001 = XT, crystal or resonator connected to GP4 and GP5
 *               000 = LP, crystal or resonator connected to GP4 and GP5
 */

                 // Gg---DCBMPWXYZ
#pragma config |= 0b11000111001100


/*------------------------------------------------------------------------------
 *
 *      Some funny #defines
 */

#define FALSE     0         // boolean constants
#define TRUE      1


/*------------------------------------------------------------------------------
 *
 *      EEPROM map
 */

// EEPROM not used...


/*------------------------------------------------------------------------------
 *
 *      I/O pins
 *
 *        GP5 - Output, sidetone speaker, via 22u + 220ohm + 10k pot to speaker
 *        GP4 - Output, transmitter keying, active high, via transistor to radio
 *        GP3 - Input, ACS switch, 1=ACS on, external pull-up
 *        GP2 - Analog input, 10k speed potentiometer, 0..VDD => 25..250 CPM
 *        GP1 - Input, dash paddle, active low, internal pull-up
 *        GP0 - Input, dot paddle, active low, internal pull-up
 */

//      Pin access macros:

#define oTONE     GPIO.5
#define oKEY      GPIO.4
#define iSW_ACS   GPIO.3

#define DOTBIT      0
#define DASHBIT     1
#define DOTMASK     (1 << DOTBIT)
#define DASHMASK    (1 << DASHBIT)
#define PADDLEMASK  (DOTMASK | DASHMASK)

//      Port state at power-up:

#define INIPORT 0b000000
#define INITRIS 0b001111


/*------------------------------------------------------------------------------
 *
 *      Global variables
 *
 *      PIC12F630/675 memory map:
 *
 *      0x20-0x5F/0xA0-0xDF, 64 bytes mapped in both banks
 */

uns16 wSpeed;             // transfers CW speed from main loop to ISR
bit qNewSpeedAvail;       // sync flag for atomic CW speed transfer from main loop to ISR


/*------------------------------------------------------------------------------
 *
 *      Interrupt service routine - 256us TMR0 tick (3906.25Hz)
 */

#pragma origin 4

interrupt vIntServer (void)
{
  static bit qKey;
  static char bState;
  static char bSideToneDiv;
  static char bCheckMask;
  static char bPaddleLatch;
  static uns16 wCwTimer;
  static uns16 wCwTimeUnit;
  bit qCwTmrExpired;
  char bSavePLH;

  int_save_registers              // save W and STATUS

  T0IF = 0;                       // clear TMR0 interrupt flag

  oKEY = qKey;                    // copy key flag to hardware here for EXACT timing ;-)

  bSideToneDiv++;
  if ((bSideToneDiv & 3) == 0)    // divide down to 488 Hz sidetone
  {
    if (qKey)
      oTONE = !oTONE;             // toggle sidetone output pin
    else
      oTONE = 0;                  // keep it low to save current
  }

  bPaddleLatch |= ~GPIO;          // sample paddle inputs and...
  bPaddleLatch &= PADDLEMASK;     // ...remember if any paddle was touched

  qCwTmrExpired = FALSE;          // run CW timer
  if (wCwTimer)
    if (--wCwTimer == 0)
      qCwTmrExpired = TRUE;

  if (qNewSpeedAvail)
  {
    wCwTimeUnit = wSpeed;         // two-byte transfer, must be atomic
    qNewSpeedAvail = FALSE;
  }

  bSavePLH = PCLATH;

  bState &= 3;      // SM6LKM's "what if the jump table index gets corrupted" paranoia :-)
  skip (bState);                  // skip is faster than switch()...
  goto st0;
  goto st1;
  goto st2;
  goto st3;

#pragma computedGoto 0


st0:                              // idle state, wait for paddle activity
  if (bPaddleLatch)               // any paddle pressed?
  {
st0_keydown:
    qKey = TRUE;                  // yes, set key-down flag
    wCwTimer = wCwTimeUnit;       // assume dot, load 1T and...

    if (bPaddleLatch.DASHBIT)     // dash paddle pressed?
    {
      wCwTimer *= 2;              // yes, forget about dot, make it 3T
      wCwTimer += wCwTimeUnit;
      bCheckMask = DOTMASK;       // check dot paddle activity while sending the dash
    }
    else
      bCheckMask = DASHMASK;      // check dash paddle activity while sending the dot
    bState++;                     // enter key-down delay state next tick
  }
  goto state_machine_done;


st1:                              // key-down delay state
  if (qCwTmrExpired)              // 1T or 3T dot/dash duration finished?
  {
    qKey = FALSE;                 // yes, clear key-down flag
    bPaddleLatch &= bCheckMask;   // forget latched "same paddle press" during key-down
    wCwTimer = wCwTimeUnit;       // load 1T for inter-element space
    bState++;                     // enter inter-element space delay state next tick
  }
  goto state_machine_done;


st2:                              // inter-element space delay state
  if (qCwTmrExpired)
  {
    if (bPaddleLatch & bCheckMask)  // opposite paddle pressed since current element started?
      bPaddleLatch &= bCheckMask; // yes, give it priority over current element type
    if (bPaddleLatch)             // any kind of element waiting to be sent?
    {
      bState = 0;                 // yes, jump back to square one and...
      goto st0_keydown;           // ...send it immediately
    }

    if (iSW_ACS)                  // automatic character space enabled?
    {
      wCwTimer = 2 * wCwTimeUnit; // yes, add 2T for a total of 3T inter-character space
      bCheckMask = 0;             // used to catch the *first* paddle touch during the delay
      bState++;                   // enter inter-character space delay state next tick
    }
    else
      bState = 0;                 // no ACS, enter idle state next tick

  }
  goto state_machine_done;


st3:                              // inter-character space delay state
  if (bCheckMask == 0)            // set flag for the first pressed paddle *ONLY*, if any
  {
    if (bPaddleLatch.DASHBIT)     // dash pressed during inter-character delay?
      bCheckMask.DASHBIT = TRUE;
    else
    if (bPaddleLatch.DOTBIT)      // dot pressed during inter-character delay?
      bCheckMask.DOTBIT = TRUE;
  }
  if (qCwTmrExpired)
  {
    bState = 0;                   // done, we want to go back to state 0 in any case
    if (bCheckMask)               // any paddling while doing inter-character space delay?
    {
      bPaddleLatch = bCheckMask;  // make state 0 respond to saved "first paddle"
      goto st0_keydown;           // immediately!
    }
  }
state_machine_done:

  PCLATH = bSavePLH;
  int_restore_registers           // restore W and STATUS
}


/*------------------------------------------------------------------------------
 *
 *      Speed potentiometer analog measurement
 *
 *      Input:
 *        Voltage on GP2 pin, 0..Vdd
 *
 *      Return value:
 *        Number of 256us ticks per CW time unit.
 */

uns16 wCwSpeedPot (void)
{
  uns16 wTmp, wTicksPerCWU;

  ADCON0 = 0b1.0.00.10.0.1;   // Right justified, Vref=Vdd, AN2=GP2, ADC on
  nop2 ();                    // 10 us acquisition/sampling delay
  nop2 ();
  nop2 ();
  nop2 ();
  nop2 ();
  GO = TRUE;                  // start conversion
  while (GO)                  // wait for EOC
    ;
  ADON = FALSE;               // turn off ADC to save a few uA :-)

  wTmp.high8 = ADRESH;
  wTmp.low8 = ADRESL;

  //                         6
  //  Dot length = --------------------- [seconds]
  //                "PARIS Speed" [CPM]
  //
  //  ADC 0..1023 => 25..250 characters/minute, linear scale, 0.22 CPM resolution:
  //
  //                                                 639375
  //  Number of 256us ticks per CW time unit = ------------------- (exactly! ;-)
  //                                            2 * (3*ADC + 341)

  wTmp *= 3;
  wTmp += 341;
  wTmp *= 2;
  wTicksPerCWU = 639375 / wTmp;   // 24-bit div used here! Aaaarrrgh!
  return wTicksPerCWU;
}


/*------------------------------------------------------------------------------
 *
 *      Initialize/re-initialize critical registers that
 *      may change state due to ESD, feed the dog too.
 */

void vSetPinModes (void)
{
  TRISIO = INITRIS;           // set pin directions
  CMCON = 0b00000111;         // comparator off
  VRCON = 0;                  // Vref module off
  WPU = 0b000011;             // internal pull-ups enabled on paddle inputs
  IOCB = 0b000000;            // no pin change wake-up
  ANSEL = 0b00010100;         // A/D clock = Fosc/8, GP2 is analog pin

  // OPTION register bits:
  //
  // Bit 7 = /GPPU  = 0, internal pull-up globally enabled
  // Bit 6 = INTEDG = 1, RB0 interrupt on rising edge (not used)
  // Bit 5 = T0CS   = 0, internal TMR0 clock
  // Bit 4 = T0SE   = 1, falling edge on GP2 increments TMR0 (not used)
  // Bit 3 = PSA    = 1, prescaler assigned to WDT (not TMR0)
  // Bit 2..0 = PS2..PS0 = 111, WDT prescaler = 128, WDT timeout ~2.3s
  //                            TMR0 overflow rate = 3906.25 Hz

  clrwdt ();
  OPTION = 0b01011111;
}


/*------------------------------------------------------------------------------
 *
 *      Main program
 */

void main (void)
{
  GPIO = INIPORT;                       // set initial output pin levels

  vSetPinModes ();                      // set pin directions, feed the WDT etc.

  // Calibrate internal RC oscillator:

  #asm
    DW 0x23FF;                          // CALL 0x3FF, CALVAL => W
  #endasm
  OSCCAL = W;

  // Initialize variables:

  clearRAM ();                          // clear all RAM

  // Start the realtime tick

  T0IE = 1;    // enable TMR0 overflow interrupt
  GIE = 1;     // go!

  // Main loop

  for (;;)
  {
    vSetPinModes ();                    // keep the PIC in good shape
    if (qNewSpeedAvail == FALSE)        // has the ISR already swallowed the speed value?
    {
      wSpeed = wCwSpeedPot ();          // yes, get another one
      qNewSpeedAvail = TRUE;            // tell ISR that a new (atomic) speed value is ready
    }
  }
}
